from typing import Sequence

import numpy as np
import pytest

from anomed_utils import arrays


@pytest.fixture
def empty_ndarray():
    return np.array([])


@pytest.fixture
def zero_confusion_matrix():
    return dict(tp=0, tn=0, fp=0, fn=0)


@pytest.fixture
def confusion_matrix_example_arrays():
    return dict(
        gt=np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 0], dtype=np.bool_),
        pred=np.array([0, 1, 0, 1, 0, 1, 0, 0, 1, 1], dtype=np.bool_),
    )


@pytest.fixture()
def shuffle_ndarray():
    return np.arange(15)


@pytest.fixture()
def shuffled_ndarray() -> dict:
    return dict(
        seed=42,
        arr=np.array([6, 7, 9, 3, 0, 11, 14, 10, 5, 2, 4, 12, 1, 13, 8]),
    )


def test_binary_confusion_matrix_empty(empty_ndarray, zero_confusion_matrix):
    assert (
        arrays.binary_confusion_matrix(empty_ndarray, empty_ndarray)
        == zero_confusion_matrix
    ), "Should be confusion matrix of only zeroes."


def test_binary_confusion_matrix(confusion_matrix_example_arrays):
    gt = confusion_matrix_example_arrays["gt"]
    pred = confusion_matrix_example_arrays["pred"]
    expected_result = dict(tp=3, tn=4, fp=2, fn=1)
    assert (
        arrays.binary_confusion_matrix(prediction=pred, ground_truth=gt)
        == expected_result
    ), f"Should be {expected_result}"


def test_binary_confusion_matrix_improper_args(confusion_matrix_example_arrays):
    gt = confusion_matrix_example_arrays["gt"].astype(np.uint8)
    pred = confusion_matrix_example_arrays["pred"].astype(np.uint8)
    with pytest.raises(ValueError):
        arrays.binary_confusion_matrix(prediction=pred, ground_truth=gt)


def test_shuffles_empty(empty_ndarray):
    assert arrays.shuffles(empty_ndarray) == []
    expectation = 3 * [empty_ndarray]
    result = arrays.shuffles(3 * [empty_ndarray])
    assert np.array_equal(expectation, result), "Should be list of empty arrays."
    result = arrays.shuffles(3 * [empty_ndarray], seed=42)
    assert np.array_equal(
        expectation, result
    ), "Should be list of empty arrays, even with explicit seed."


def _arrays_equal(arrays: Sequence[np.ndarray]) -> bool:
    if len(arrays) == 0:
        return True
    else:
        reference_array = arrays[0]
        return all([np.array_equal(reference_array, arr) for arr in arrays[1:]])


def test_shuffles_equal_shuffling(shuffle_ndarray):
    shuffled = arrays.shuffles(3 * [shuffle_ndarray])
    assert _arrays_equal(shuffled), "All arrays should be shuffled equally"
    shuffled = arrays.shuffles(3 * [shuffle_ndarray], seed=42)
    assert _arrays_equal(
        shuffled
    ), "All arrays should be shuffled equally, even with explicit seed."


def test_shuffles_actually_shuffling(shuffle_ndarray, shuffled_ndarray):
    [shuffled_array] = arrays.shuffles([shuffle_ndarray], seed=shuffled_ndarray["seed"])
    assert np.array_equal(
        shuffled_ndarray["arr"],
        shuffled_array,
    ), "Shuffling should actually change the input array."


def test_random_partitions_empty(empty_ndarray):
    [(part1, part2)] = arrays.random_partitions(
        [empty_ndarray], total_length=0, desired_length=0
    )
    assert part1.size == 0 and part2.size == 0, "Both parts should be empty."


def test_random_partitions_sizes(shuffle_ndarray):
    n = len(shuffle_ndarray)
    partitions = arrays.random_partitions(
        3 * [shuffle_ndarray], total_length=n, desired_length=5
    )
    for partition in partitions:
        (part1, part2) = partition
        assert len(part1) == 5 and len(part2) == n - 5


def test_random_partitions_actually_partitioning(shuffle_ndarray):
    expected_part1 = np.array([6, 7, 9, 3, 0, 11, 14])
    expected_part2 = np.array([10, 5, 2, 4, 12, 1, 13, 8])
    [(part1, part2)] = arrays.random_partitions(
        [shuffle_ndarray], total_length=15, desired_length=7, seed=42
    )
    assert np.array_equal(expected_part1, part1) and np.array_equal(
        expected_part2, part2
    )


def test_binary_confusion_matrix_raising():
    with pytest.raises(ValueError):
        arrays.binary_confusion_matrix(
            np.ones(shape=(5,), dtype=np.bool_), np.ones(shape=(10,), dtype=np.bool_)
        )



import falcon
import falcon.app
import numpy as np
import pytest
from falcon import testing

from anomed_utils import web


@pytest.fixture()
def empty_array():
    return np.array([])


@pytest.fixture()
def int_array():
    return np.arange(10)


@pytest.fixture()
def float_array():
    return np.arange(10) + 0.5


@pytest.fixture()
def object_array():
    return np.array(3 * ["foo", "bar", "baz"])


@pytest.fixture()
def client():
    app = falcon.App()
    app.add_route("/", web.StaticJSONResource(dict(message="hello world")))
    return testing.TestClient(app=app)


def test_StaticJSONResource(client):
    json = dict(message="hello world")
    response = client.simulate_get("/")
    assert response.json == json


def test_array_to_bytes_conversion(empty_array, int_array, float_array, object_array):
    expected_result = dict(
        empty=empty_array, ints=int_array, floats=float_array, objs=object_array
    )
    result = web.bytes_to_named_ndarrays(web.named_ndarrays_to_bytes(expected_result))
    assert expected_result.keys() == result.keys()
    for key in expected_result.keys():
        assert np.array_equal(expected_result[key], result[key])



import json
import logging
from io import BytesIO
from pathlib import Path
from typing import Any, Callable, Iterable

import falcon
import numpy as np
import requests
from filelock import FileLock, Timeout


class StaticJSONResource:
    """Any JSON serializable object, representing a "static" resource (i.e. a
    resource that does not depend on request parameters).

    The object will be represented as a plain JSON string, when a GET request is
    invoked."""

    def __init__(self, obj: Any):
        """
        Parameters
        ----------
        obj : Any
            A JSON serializable object, i.e. is should be compatible with
            `json.dumps`.
        """
        self._obj = obj

    def on_get(self, _, resp: falcon.Response):
        resp.text = json.dumps(self._obj)


class FitResource:
    """This resource is intended for fitting estimator following the supervised
    learning paradigm.
    """

    def __init__(
        self,
        data_getter: Callable[[], dict[str, np.ndarray]],
        model: Any,
        model_filepath: str | Path,
    ) -> None:
        """
        Parameters
        ----------
        data_getter : Callable[[], dict[str, np.ndarray]]
            A callable which will provide fitting data, suitable for the
            `model`'s `fit` method. It is assumed, that the resulting data is
            compatible with `model`. No further validation is performed.
            If data_getter fails to obtain the fitting data, it is expected to
            raise an instance of `falcon.HTTPError` (or a subclass).
        model : Any
            The model/estimator to fit (supervised learning paradigm).
        model_filepath : str | Path
            Where to persist the fitted model.
        """
        self._get_fit_data = data_getter
        self._model = model
        self._model_filepath = Path(model_filepath)
        self._model_lock = FileLock(
            self._model_filepath.with_suffix(".lock"), blocking=False
        )

    def on_post(self, req: falcon.Request, resp: falcon.Response) -> None:
        """Initiate the fitting process for `self.model`.

        Parameters
        ----------
        req : falcon.Request
            A Falcon request object (irrelevant).
        resp : falcon.Response
            A Falcon response object. On success, the status code will be set to
            201 and a meaningful, JSON encoded message will be given back.

        Raises
        ------
        falcon.HTTPServiceUnavailable
            If fitting is already in progress (no parallel fitting).
        falcon.HTTPError
            If obtaining training data failed.
        """
        logging.info("Obtaining fitting data.")
        fit_data = self._get_fit_data()
        try:
            with self._model_lock:
                logging.info("Initiating fitting.")
                self._model.fit(**fit_data)
                logging.info("Persisting fitted model.")
                self._model.save(self._model_filepath)
        except Timeout:
            raise falcon.HTTPServiceUnavailable(
                description="Fitting is already in progress. Aborting current attempt.",
            )
        success_message = "Fitting has been completed successfully."
        logging.info(success_message)
        resp.status = falcon.HTTP_CREATED
        resp.text = json.dumps(dict(message=success_message))


def named_ndarrays_to_bytes(named_arrays: dict[str, np.ndarray]) -> bytes:
    """Convert named NumPy arrays to a compressed bytes sequence.

    Use this for example as payload data in a POST request.

    Parameters
    ----------
    named_arrays : dict[str, np.ndarray]
        The named NumPy arrays.

    Returns
    -------
    bytes
        A compressed bytes sequence.

    Notes
    -----
    This is the inverse to `bytes_to_named_ndarrays`.
    """
    compressed_arrays = BytesIO()
    np.savez_compressed(compressed_arrays, **named_arrays)
    return compressed_arrays.getvalue()


def bytes_to_named_ndarrays(data: bytes) -> dict[str, np.ndarray]:
    """Convert a bytes sequence of named (and compressed) NumPy arrays back to
    arrays.

    Use this for example to retrieve NumPy arrays from an HTTP response.

    Parameters
    ----------
    data : bytes
        The bytes representation of a (compressed)

    Returns
    -------
    dict[str, np.ndarray]
        The named arrays.

    Raises
    ------
    OSError
        If the input file does not exist or cannot be read.
    ValueError
        If the file contains an object array.


    Notes
    -----
    This in the inverse to `named_ndarrays_to_bytes`.
    """
    arrays = np.load(BytesIO(data))
    return {name: arrays[name] for name in arrays.files}


def bytes_to_named_ndarrays_or_raise(
    payload: bytes,
    expected_array_labels: Iterable[str],
    error_status: str | int,
    error_message: str | None = None,
) -> dict[str, np.ndarray]:
    """This is basically a wrapper around `bytes_to_named_ndarrays` which raises
    `falcon.HTTPError` if `bytes_to_named_ndarrays` would fail decoding the
    byte string.

    Use this as building block for falcon-based web servers.

    Parameters
    ----------
    payload : bytes
        The byte string which presumably contain named NumPy arrays.
    expected_array_labels : Iterable[str]
        Which arrays are expected (identified by their name) to be present. If
        at least one of them is missing, raise an exception.
    error_status : str | int
        Which error status to use if an exception is raised.
    error_message : str | None, optional
        A more detailed error message to provide, if an exception is raised. Use
        a generic one that suits `error_status`, if `None`.

    Returns
    -------
    dict[str, np.ndarray]
        Named NumPy arrays.

    Raises
    ------
    falcon.HTTPError
        If parsing fails.
    """
    try:
        arrays = bytes_to_named_ndarrays(payload)
        if not all(
            [expected_label in arrays for expected_label in expected_array_labels]
        ):
            raise ValueError("Array payload does not contain all expected labels.")
    except (OSError, ValueError, EOFError):
        if error_message is None:
            error_message = "Array payload parsing (or validation) failed."
        logging.exception(error_message)
        raise falcon.HTTPError(status=error_status, description=error_message)
    return arrays


def get_named_arrays_or_raise(
    data_url: str,
    expected_array_labels: Iterable[str],
    params: dict[str, Any] | None = None,
    timeout: float | None = None,
) -> dict[str, np.ndarray]:
    """Obtain byte-serialized NumPy arrays (e.g. via `named_ndarrays_to_bytes`)
    from a remote location via GET request and validate the payload. Raise a
    `falcon.HTTPError` if it fails.

    Parameters
    ----------
    data_url : str
        The remote location to access via GET request. It should respond
        byte-serialized named NumPy arrays.
    expected_array_labels : Iterable[str]
        Which arrays (identified by name) the payload should carry.
    params : dict[str, Any] | None, optional
        Optional request parameters to provide. By default `None` (no
        parameters).
    timeout : float | None, optional
        The timeout to use for the connection to `data_url`. By default `None`
        (wait until connection is closed).

    Returns
    -------
    dict[str, np.ndarray]
        Named NumPy arrays.

    Raises
    ------
    falcon.HTTPServiceUnavailable
        If the connection to, or the response from the remote location was
        faulty.
    falcon.HTTPInternalServerError
        If parsing (named) NumPy arrays from the requested bytes payload failed.
    """
    try:
        data_resp = requests.get(url=data_url, params=params, timeout=timeout)
        if data_resp.status_code != 200:
            raise ValueError()
    except (requests.RequestException, ValueError):
        error_msg = "Unable to obtain data from remote location (timeout or error)."
        logging.exception(error_msg)
        raise falcon.HTTPServiceUnavailable(description=error_msg)
    arrays = bytes_to_named_ndarrays_or_raise(
        data_resp.content,
        expected_array_labels=expected_array_labels,
        error_status=falcon.HTTP_INTERNAL_SERVER_ERROR,
        error_message="Array payload validation failed.",
    )
    return arrays



from typing import Iterable, Sequence

import numpy as np


def random_partitions(
    arrays: Iterable[np.ndarray],
    total_length: int,
    desired_length: int,
    seed: int | None = None,
) -> list[tuple[np.ndarray, np.ndarray]]:
    """Partition an iterable of NumPy arrays, i.e. split each of them into two
    subarrays, where each element of the original array is either in the first,
    or second subarray (not both and not neither).

    The subarrays are also shuffled, and the way they are shuffled (i.e. how the
    indices of the original elements move around in the partition) stays the
    same for all arrays of the iterable.

    For example, these are some ways to partition a single array:
    ```
        [1, 2, 3, 4] -> ([1, 2], [3, 4])
        [1, 2, 3, 4] -> ([2, 4], [3, 1])
        [1, 2, 3, 4] -> ([2], [3, 1, 4])
        [1, 2, 3, 4] -> ([], [3, 1, 4, 2])
        [1, 1, 2, 2] -> ([1, 2, 1], [2])
    ```
    And this is how the shuffling extends over multiple arrays:
    ```
        [1, 2, 3, 4], [5, 6, 7, 8] -> ([4, 3], [1, 2]), ([8, 7], [5, 6])
        [1, 2, 3, 4], [5, 6, 7, 8] -> ([3], [4, 1, 2]), ([7], [8, 5, 6])
    ```

    Parameters
    ----------
    arrays : Iterable[np.ndarray]
        An iterable of NumPy arrays, assuming all to have the same length (first
        dimension), but not necessarily the same shape or dtype.
    total_length : int
        The (total) length of each array in `arrays`.
    desired_length : int
        The desired size of the first components. The second components contain
        the remaining entries.
    seed : int | None, optional
        The seed to use for randomness. By default `None` (obtaining randomness
        at runtime).

    Returns
    -------
    list[tuple[np.ndarray, np.ndarray]]
        The list of partitions (corresponding to the arrays in `arrays`).
    """
    rng = np.random.default_rng(seed)
    indices = np.arange(total_length)
    rng.shuffle(indices)
    keep = indices[:desired_length]
    leave = indices[desired_length:]
    splits = [(anon_arr[keep], anon_arr[leave]) for anon_arr in arrays]
    return splits


def shuffles(arrays: Sequence[np.ndarray], seed: int | None = None) -> list[np.ndarray]:
    """Shuffle each of the arrays in the same way, meaning the way the first
    array is permuted is equal among all other arrays.

    Parameters
    ----------
    arrays : Sequence[np.ndarray]
        A sequence of NumPy arrays, assuming all to have the same length (first
        dimension), but not necessarily the same shape or dtype.
    seed : int | None, optional
        The seed to use for randomness. By default `None` (obtaining randomness
        at runtime).

    Returns
    -------
    list[np.ndarray]
        A list of shuffled arrays.
    """
    if len(arrays) == 0:
        return []
    else:
        n = len(arrays[0])
        shuffled_arrays = [
            shuffled_array
            for (shuffled_array, _) in random_partitions(arrays, n, n, seed)
        ]
        return shuffled_arrays


def binary_confusion_matrix(
    prediction: np.ndarray, ground_truth: np.ndarray
) -> dict[str, int]:
    """Calculate the true positives, true negatives, false positives and false
    negatives of a binary (boolean) prediction compared to a binary ground
    truth.

    Parameters
    ----------
    prediction : np.ndarray
        The binary (boolean) prediction.
    ground_truth : np.ndarray
        The binary (boolean) ground truth.

    Returns
    -------
    dict[str, int]
        A dictionary with keys "tp", "tn", "fp" and "fn".

    Raises
    ------
    ValueError
        If the arguments are not of dtype `np.bool_`, or if their shape does not
        match.
    """
    if len(prediction) == 0 or len(ground_truth) == 0:
        return dict(tp=0, tn=0, fp=0, fn=0)
    if not prediction.dtype == ground_truth.dtype == np.bool_:
        raise ValueError(
            f"Prediction's dtype ({prediction.dtype}) or ground_truth's "
            f"dtype({ground_truth.dtype}) is not `np.bool`."
        )
    if not prediction.shape == ground_truth.shape:
        raise ValueError(
            f"Shape mismatch of prediction {prediction.shape} and ground_truth "
            f"{ground_truth.shape}"
        )
    _pred = prediction.astype(np.int8)
    _gt = ground_truth.astype(np.int8)

    result = np.zeros((2, 2), dtype=int)
    for i in range(len(_pred)):
        result[_gt[i], _pred[i]] += 1

    return dict(tp=result[1, 1], tn=result[0, 0], fp=result[0, 1], fn=result[1, 0])
